#Copyright (c) 2010, Pascal Widdershoven (for all parts of the original protocol_handler.py)
#Copyright (c) 2012, mightbemighty (mightbemighty@gmail.com) (for all the rest)

import logging, appletv, re, lib.biplist

from twisted.internet import reactor
from twisted.internet.error import CannotListenError
from twisted.web import resource, server, http

import mimetools
from StringIO import StringIO

log = logging.getLogger('openair')

from mediabackends.enigma2_media_backend import AirMoviePlayer

__version__ = "0.1"

class AirplayProtocolHandler(object):

	def __init__(self, port, media_backend):
		self._media_backend = media_backend
		self._media_backend.onPlayStateChanged.append(self._onPlayStateChanged)
		self._port = port
		self.reverseRequests = {}

	def _onPlayStateChanged(self, state):
		log.debug("state=%s", state)
		eventmap = {
			AirMoviePlayer.STATE_STOPPED : appletv.EVENT_INFO_STOPPED,
			AirMoviePlayer.STATE_PLAYING : appletv.EVENT_INFO_PLAYING,
			AirMoviePlayer.STATE_PAUSED  : appletv.EVENT_INFO_PAUSED,
		}
		self.sendEventInfo(eventmap[state])

	def setReverseRequest(self, sessionId, request):
		log.debug("setReverseRequest: sessionId=%s", sessionId)
		self.reverseRequests[sessionId] = request

	def sendEventInfo(self, event, sessionId='dummy'):
		log.info("[OpenAir] REVERSE /event - event=%s", event)
		header_template = "POST /event HTTP/1.1\r\nContent-Type: text/x-apple-plist+xml\r\nContent-Length: %s\r\nx-apple-session-id: %s\r\n"

		request = self.getReverseRequest(sessionId)
		if request != None:
			xml = appletv.EVENT_INFO %event
			header = header_template %(len(xml), sessionId)
			req = "%s\r\n%s\r\n\r\n" %(header, xml)
			log.debug("sendEventInfo req=\n%s", req)

			request.write(req)

	def getReverseRequest(self, sessionId):
		request = self.reverseRequests.get(sessionId, None)
		log.debug("getReverseRequest: sessionId=%s, request=%s", sessionId, request)
		return request

	def start(self):
		handlers = {
			'reverse' : AirplayProtocolHandler.ReverseHandler,
			'play' : AirplayProtocolHandler.PlayHandler,
			'scrub' : AirplayProtocolHandler.ScrubHandler,
			'rate' : AirplayProtocolHandler.RateHandler,
			'photo' : AirplayProtocolHandler.PhotoHandler,
			'authorize' : AirplayProtocolHandler.AuthorizeHandler,
			'server-info' : AirplayProtocolHandler.ServerInfoHandler,
			'slideshow-features' : AirplayProtocolHandler.SlideshowFeaturesHandler,
			'playback-info' : AirplayProtocolHandler.PlaybackInfoHandler,
			'stop' : AirplayProtocolHandler.StopHandler,
		}

		root = AirplayProtocolHandler.Root()
		log.info('[OpenAir] AirplayProtocolHandler:: starting')
		for path, handler in handlers.iteritems():
			log.info('[OpenAir] AirplayProtocolHandler:: Adding %s', path)
			child = handler(self._media_backend, self.setReverseRequest, self.sendEventInfo)
			root.putChild(path, child)

		try:
			factory = server.Site(root)
			self._deferred = reactor.listenTCP(self._port, factory, interface='0.0.0.0')
		except CannotListenError, e:
			log.error(str(e))
		finally:
			log.debug('Cleaning up')
			self._media_backend.cleanup()

	def stop(self):
		try:
			self._deferred.stopListening()
		except:
			pass

	class Root(resource.Resource):
		isLeaf = False
		def render_GET(self, request):
			return "[OpenAir] for enigma2, version: %s" %(__version__)

		def render_POST(self, request):
			return self.render_GET(request)

		def render_PUT(self, request):
			return self.render_GET(request)

	class BaseHandler(resource.Resource):
		"""
		Base request handler, all other handlers should inherit from this class.

		Provides some logging and media backend assignment.
		"""
		isLeaf = True
		def __init__(self, media_backend, setReverseRequestCB, sendEventInfoCB):
			self._media_backend = media_backend
			self.sendEventInfo = sendEventInfoCB
			self.setReverseRequest = setReverseRequestCB

			self.request = None
			self._session_id = None
			self.body = None
			self._finished = False

		def render(self, request):
			log.info("[OpenAir] path=%s" %request.path)
			log.debug("[OpenAir] headers=%s", request.received_headers)

			self.request = request
			self._session_id = self.request.getHeader("x-apple-session-id")
			if self._session_id == None:
				self._session_id = 'dummy'

			self._finished = False
			self.prepare()

			self.request.responseHeaders.removeHeader('Content-Type')
			self.request.responseHeaders.removeHeader('Server')
			self.set_header("User-Agent",  "AppleCoreMedia/1.0.0.8F455 (AppleTV; U; CPU OS 4_3 like Mac OS X; de_de)")
			self.set_header("Content-Length", "0")

			resource.Resource.render(self, request)
			self._finish()
			return server.NOT_DONE_YET

		def render_GET(self, request):
			self.body = None
			self.get()

		def render_POST(self, request):
			self.body = request.content.getvalue()
			log.debug("body=%s" %self.body)
			self.post()

		def render_PUT(self, request):
			self.body = request.content.read()
			log.debug("body size is %s bytes" %len(self.body))
			self.put()

		def prepare(self):
			pass

		def get(self):
			print "GET not implemented for %s" %(self.request.path)

		def post(self):
			print "POST not implemented for %s" %(self.request.path)

		def put(self):
			print "PUT not implemented for %s" %(self.request.path)

		def set_status(self, code):
			self.request.setResponseCode(code)

		def set_header(self, header, value):
			self.request.setHeader(header, value)

		def write(self, content):
			self.set_header('content-length', len(content))
			self.request.write(content)

		def finish(self):
			pass

		def _finish(self):
			if not self._finished:
				self._finished = True
				self.request.finish()

	class ReverseHandler(BaseHandler):
		"""
		Handler for /reverse requests.

		The reverse command is the first command sent by Airplay,
		it's a handshake.
		"""

		def post(self):
			self.set_status(101)
			self.set_header('Upgrade', 'PTTH/1.0')
			self.set_header('Connection', 'Upgrade')
			self.setReverseRequest(self._session_id, self.request);

		def _finish(self):
			self.request.write('\r\n\r\n')

	class PlayHandler(BaseHandler):
		"""
		Handler for /play requests.

		Contains a header like format in the request body which should contain a
		Content-Location and optionally a Start-Position.
		"""

		def post(self):
			"""
			Immediately finish this request, no need for the client to wait for
			backend communication.
			"""
#			self.sendEventInfo(appletv.EVENT_INFO_LOADING, self._session_id)
			self.finish()

			if self.request.getHeader('Content-Type') == 'application/x-apple-binary-plist':
				body = lib.biplist.readPlistFromString(self.body)
			else:
				body = mimetools.Message(StringIO(self.body))

			print "[PlayHandler] body=%s" %body
			if 'Content-Location' in body:
				position_percentage = 0
				if 'Start-Position' in body:
					"""
					Airplay sends start-position in percentage from 0 to 1.
					Media backends expect a percentage from 0 to 100.
					"""
					try:
						str_pos = str(body['Start-Position'])
					except ValueError:
						log.warning('Invalid start-position supplied: ', str_pos)
					else:
						position_percentage = float(str_pos.replace(",", ".")) * 100

			url = body['Content-Location']
			log.debug('Playing %s', url)
			self._media_backend.play_movie(url, position_percentage)

	class ScrubHandler(BaseHandler):
		"""
		Handler for /scrub requests.

		Used to perform seeking (POST request) and to retrieve current player position (GET request).
		"""

		def get(self):
			"""
			Will return None, None if no media is playing or an error occures.
			"""
			position, duration = self._media_backend.get_player_position()

			"""
			Should None values be returned just default to 0 values.
			"""
			if not position:
				duration = position = 0

			body = 'duration: %f\r\nposition: %f\r\n' % (duration, position)
			self.set_header("content-type", "text/parameters")
			self.write(body)

		def post(self):
			"""
			Immediately finish this request, no need for the client to wait for
			backend communication.
			"""
			self.finish()

			if 'position' in self.request.args:
				try:
					str_pos = self.request.args['position'][0]
					position = int(float(str_pos.replace(",", ".")))
				except ValueError:
					log.warn('Invalid scrub value supplied: %s', str_pos)
				else:
					self._media_backend.set_player_position(position)

	class RateHandler(BaseHandler):
		"""
		Handler for /rate requests.

		The rate command is used to play/pause media.
		A value argument should be supplied which indicates media should be played or paused.

		0.000000 => pause
		1.000000 => play
		"""

		def post(self):
			"""
			Immediately finish this request, no need for the client to wait for
			backend communication.
			"""
			self.finish()

			if 'value' in self.request.args:
				play = bool(float(self.request.args['value'][0]))

				if play:
					self._media_backend.play()
				else:
					self._media_backend.pause()

	class PhotoHandler(BaseHandler):
		"""
		Handler for /photo requests.

		RAW JPEG data is contained in the request body.
		"""

		def put(self):
			"""
			Immediately finish this request, no need for the client to wait for
			backend communication.
			"""
			self.finish()

			if self.body:
				self._media_backend.show_picture(self.body)

	class AuthorizeHandler(BaseHandler):
		"""
		Handler for /authorize requests.

		This is used to handle DRM authorization.
		We currently don't support DRM protected media.
		"""

		def prepare(self):
			log.warning('Trying to play DRM protected, this is currently unsupported.')
			log.debug('Got an authorize %s request', self.request.method)
			log.debug('Authorize request info: %s %s %s', self.request.requestHeaders, self.request.args, self.body)

		def get(self):
			pass

		def post(self):
			pass

	class StopHandler(BaseHandler):
		"""
		Handler for /stop requests.

		Sent when media playback should be stopped.
		"""

		def post(self):
			self._media_backend.stop_playing()
			self.sendEventInfo(appletv.EVENT_INFO_STOPPED, self._session_id)

	class ServerInfoHandler(BaseHandler):
		"""
		Handler for /server-info requests.

		Usage currently unknown.
		Available from IOS 4.3.
		"""

		def get(self):
			self.set_header('Content-Type', 'text/x-apple-plist+xml')
			self.write(appletv.SERVER_INFO)

	class SlideshowFeaturesHandler(BaseHandler):
		"""
		Handler for /slideshow-features requests.

		Usage currently unknown.
		Available from IOS 4.3.
		"""

		def get(self):
			"""
			I think slideshow effects should be implemented by the Airplay device.
			The currently supported media backends do not support this.

			We'll just ignore this request, that'll enable the simple slideshow without effects.
			"""
			pass

	class PlaybackInfoHandler(BaseHandler):
		"""
		Handler for /playback-info requests.
		"""

		def get(self):
			playing = self._media_backend.is_playing()
			position, duration = self._media_backend.get_player_position()

			if position < 0:
				position = duration = 0
			else:
				position = float(position)
				duration = float(duration)

			body = appletv.PLAYBACK_INFO % (duration, duration, position, int(playing), duration)
			log.debug("[PlaybackInfoHandler].get: duration=%s, position=%s, playing=%s", duration, position, playing)

			self.set_header('Content-Type', 'text/x-apple-plist+xml')
			self.write(body)
